"""Settings for Veliberator"""
import os
from ConfigParser import SafeConfigParser

XML_SECTION = 'XML'
STATION_SECTION = 'STATION'
DATABASE_SECTION = 'DATABASE'
PROXY_SECTION = 'PROXY'

ALL_SECTIONS = [XML_SECTION, STATION_SECTION,
                DATABASE_SECTION, PROXY_SECTION]

DEFAULT_CONFIG = {
    'url_data_station': 'http://www.velib.paris.fr/service/carto',
    'url_status_station': 'http://www.velib.paris.fr/service/stationdetails/',
    'around_radius': '1200',
    'status_recent': '10',
    'almost_full': '3',
    'almost_empty': '3',
    'uri': 'sqlite:///~/.veliberator.db',
    'echo': 'False',
    'servers': ''}

config = SafeConfigParser(DEFAULT_CONFIG)

for section in ALL_SECTIONS:
    config.add_section(section)

config_read = config.read([os.path.expanduser('~/.veliberator.cfg'),
                           os.path.join(os.getcwd(), 'etc/veliberator.cfg')])

XML_URL_DATA_STATION = config.get(XML_SECTION, 'url_data_station')
XML_URL_STATUS_STATION = config.get(XML_SECTION, 'url_status_station')

STATION_AROUND_RADIUS = config.getint(STATION_SECTION, 'around_radius')
STATION_STATUS_RECENT = config.getint(STATION_SECTION, 'status_recent')
STATION_ALMOST_FULL = config.getint(STATION_SECTION, 'almost_full')
STATION_ALMOST_EMPTY = config.getint(STATION_SECTION, 'almost_empty')

DATABASE_URI = config.get(DATABASE_SECTION, 'uri')
if '~' in DATABASE_URI:
    protocol, source = DATABASE_URI.split('~')
    DATABASE_URI = '%s%s' % (protocol, os.path.expanduser('~' + source))

DATABASE_ECHO = config.getboolean(DATABASE_SECTION, 'echo')

PROXY_SERVERS = config.get(PROXY_SECTION, 'servers')
PROXY_SERVERS = PROXY_SERVERS and PROXY_SERVERS.split(';') or []

TEST_XML_URL_DATA_STATION = 'file:%s' % os.path.join(
    os.path.dirname(__file__), 'tests', 'data', 'carto_short.xml')



"""Unit tests for GeoFinder objects"""
import unittest

from veliberator.station import Station
from veliberator.geofinder import global_geofinder_cache
from veliberator.geofinder import BaseGeoFinder
from veliberator.geofinder import StationGeoFinder
from veliberator.geofinder import AddressGeoFinder
from veliberator.geofinder import pythagor_distance
from veliberator.models import StationInformation
from veliberator.cartography import Cartography
from veliberator.settings import TEST_XML_URL_DATA_STATION

STATION_AROUND_RADIUS = 1200


class BaseGeoFinderTestCase(unittest.TestCase):

    def test_ComputeSquareArea(self):
        finder = BaseGeoFinder(1, 1)
        self.assertEquals(finder.compute_square_area(STATION_AROUND_RADIUS),
                          ((0.98922021659056569, 1.0, 1.0107797834094343),
                           (0.98922021659056569, 1.0, 1.0107797834094343)))
        finder.lat = '1'
        finder.lng = '1'
        self.assertEquals(finder.compute_square_area(STATION_AROUND_RADIUS),
                          ((0.98922021659056569, 1.0, 1.0107797834094343),
                           (0.98922021659056569, 1.0, 1.0107797834094343)))
        finder.lat = 4.897645
        finder.lng = 18.798923
        self.assertEquals(finder.compute_square_area(STATION_AROUND_RADIUS),
                          ((4.8868652165905653,
                            4.8976449999999998,
                            4.9084247834094343),
                           (18.788143216590566,
                            18.798922999999998,
                            18.809702783409431)))

    def test_ComputeStationDistances(self):
        finder = BaseGeoFinder(1, 1)

        si1 = StationInformation(id=42, lat=1, lng=1)
        result = finder.compute_station_distances([si1])
        self.assertEquals(result, [])

        si2 = StationInformation(id=43, lat=4, lng=5)
        result = finder.compute_station_distances([si2])
        self.assertEquals(result, [si2])

        self.assertEquals(result[0].distance, 556217.45145026874)
        result = finder.compute_station_distances([si2], pythagor_distance)
        self.assertEquals(result[0].distance, 556597.45396636787)

        si3 = StationInformation(id=44, lat=8, lng=10)
        si4 = StationInformation(id=45, lat=6, lng=5.5)
        result = finder.compute_station_distances([si1, si2, si3, si4])
        self.assertEquals(len(result), 3)

        StationInformation.query.delete()

    def test_GetStationsInArea(self):
        test_1 = (('4.89', '4.90', '4.91'), ('18.79', '18.80', '18.81'))
        test_2 = (('48.80', '48.81', '48.82'), ('2.37', '2.38', '2.39'))

        Cartography.synchronize(TEST_XML_URL_DATA_STATION)
        finder = BaseGeoFinder(1, 1)

        self.assertEquals(len(finder.get_stations_in_area(*test_1)), 0)
        self.assertEquals(len(finder.get_stations_in_area(*test_2)), 8)

        Cartography.flush()

    def test_GetStationsAround(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)

        finder = BaseGeoFinder(1, 1)
        self.assertEquals(len(finder.get_stations_around(
            STATION_AROUND_RADIUS)), 0)

        finder.lat = 48.81
        finder.lng = 2.38
        self.assertEquals(
            [station.id for station in
             finder.get_stations_around(STATION_AROUND_RADIUS)],
            [42012, 42010, 42009, 42008, 42016, 42006, 42007, 42015])

        Cartography.flush()

    def test_Cache(self):
        key = (1, 1, STATION_AROUND_RADIUS)
        self.assertFalse(key in global_geofinder_cache)
        finder = BaseGeoFinder(key[0], key[1])
        finder.get_stations_around(key[2])
        self.assertTrue(key in global_geofinder_cache)


class StationGeoFinderTestCase(unittest.TestCase):

    def test_Init(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)

        station = Station(42008)
        geofinder = StationGeoFinder(station)
        self.assertEquals(geofinder.lat, station.informations.lat)
        self.assertEquals(geofinder.lng, station.informations.lng)

        Cartography.flush()


class AddressGeoFinderTestCase(unittest.TestCase):

    def test_Init(self):
        address = '1 place de la Bastille, 75012 Paris'
        geofinder = AddressGeoFinder(address)
        self.assertEquals(geofinder.precision, 'street_address')
        self.assertEquals(geofinder.lat, 48.8530318)
        self.assertEquals(geofinder.lng, 2.3687204)

    def test_GeoCompute(self):
        address = '1 place de la Bastille, 75012 Paris'
        geofinder = AddressGeoFinder(address)
        self.assertEquals(len(geofinder.geocompute(address)), 6)


suite = unittest.TestSuite([
    unittest.TestLoader().loadTestsFromTestCase(BaseGeoFinderTestCase),
    unittest.TestLoader().loadTestsFromTestCase(StationGeoFinderTestCase),
    unittest.TestLoader().loadTestsFromTestCase(AddressGeoFinderTestCase)])



"""Setup script for veliberator"""
import os

from setuptools import setup
from setuptools import find_packages

import veliberator

setup(
    name='veliberator',
    version=veliberator.__version__,
    license=veliberator.__license__,

    author=veliberator.__author__,
    author_email=veliberator.__email__,
    url=veliberator.__url__,

    description="Veliberator provides a Python API for getting informations "
    "about the stations of the Velib' network.",
    long_description=open(os.path.join('README.rst')).read(),
    keywords='velib, api, service',

    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],

    entry_points={
        'console_scripts': [
            'veliberator=veliberator.scripts.command_line:cmdline',
            ]
    },
    test_suite='veliberator.tests.global_test_suite',
    packages=find_packages(exclude=['tests']),

    zip_safe=False,
    platforms='any',
    include_package_data=True,

    install_requires=['SQLAlchemy==0.7.10',
                      'Elixir==0.7.1',
                      'simplejson'],
)



"""
Veliberator script
"""
import os
import sys
import gettext
from optparse import OptionParser

import sqlalchemy

import veliberator
from veliberator.station import Station
from veliberator.station import UnknowStation
from veliberator.cartography import Cartography
from veliberator.models import db_connection
from veliberator.settings import DATABASE_URI
from veliberator.settings import STATION_AROUND_RADIUS
from veliberator.geofinder import GeoFinderError
from veliberator.geofinder import AddressGeoFinder
from veliberator.models import StationInformation


translation = gettext.translation(
    'veliberator',
    os.path.join(os.path.dirname(veliberator.__file__), 'locale'))
_ = translation.gettext
__ = translation.ngettext


def synchronization():
    print _('-> Online synchronization...')
    Cartography.synchronize()
    print _('-> Synchronization complete !')


def show_status(station, distance=None):
    print _("Station '%s'") % station.id
    if distance:
        print _('%(address)s (%(distance).2fm)') % {
            'address': station.informations.full_address,
            'distance': distance}
    else:
        print station.informations.full_address

    print __('%(available)s/%(total)s bike available',
             '%(available)s/%(total)s bikes available',
             station.status.available) % {
        'available': station.status.available,
        'total': station.status.total}
    print __('%s parking place', '%s parking places',
             station.status.free) % station.status.free


def display_free_stations(stations, places, max_display):
    displayed = 0

    for station_information in stations:
        if displayed == max_display:
            break
        station = Station(station_information.id)
        if station.is_free(places):
            print '----------'
            show_status(station, station_information.distance)
            displayed += 1


class VeliberatorOptionParser(OptionParser):
    """Customized OptionParser with
    correct gettext handling"""

    def _add_help_option(self):
        self.add_option('-h', '--help', action='help',
                        help=_('Show this help message and exit'))

    def _add_version_option(self):
        self.add_option('--version', action='version',
                        help=_("Show program's version number and exit"))

    def print_help(self, file=None):
        if file is None:
            file = sys.stdout
        file.write(self.format_help())


def cmdline():
    parser = VeliberatorOptionParser(
        usage=_('%prog [station_id] [address] [options]'),
        version='%s %s' % ('%prog', veliberator.__version__))
    parser.add_option('-d', '--database', dest='database',
                      help=_('The SQLURI of the database'),
                      default=DATABASE_URI)
    parser.add_option('-p', '--places', dest='places', type='int',
                      help=_('The number of places you want'), default=1)
    parser.add_option('-m', '--max-stations', dest='max_stations', type='int',
                      help=_('The maximun stations around proposed'),
                      default=5)
    parser.add_option('--synchronize', dest='synchronize', action='store_true',
                      help=_('Only do a synchronization of the cartography'))
    (options, args) = parser.parse_args()

    print '-==* Veliberator v%s *==-' % veliberator.__version__

    try:
        db_connection(options.database)
    except sqlalchemy.exc.OperationalError:
        if options.synchronize:
            sys.exit(_('-> The database %s is unreachable.') %
                     options.database)
        else:
            print _('-> The database is unreachable, switch on RAM.')
            print _('-> Edit the configuration file, '
                    'for removing this message.')
            db_connection('sqlite://')

    if options.synchronize:
        synchronization()
        sys.exit(0)

    if not StationInformation.query.count():
        synchronization()

    if args:
        user_input = args[0]
    else:
        user_input = raw_input(_('Station ID or complete address:\n'))

    if user_input.isdigit():
        try:
            station = Station(user_input)
        except UnknowStation:
            sys.exit(_('The Station ID does not exist.'))
        show_status(station)
        if not station.is_free(options.places):
            print _('Searching on the closest stations...')
            display_free_stations(station.stations_around,
                                  options.places, options.max_stations)
    else:
        try:
            finder = AddressGeoFinder(user_input)
        except GeoFinderError:
            sys.exit(_('The provided address is not valid or imprecise.'))
        display_free_stations(
            finder.get_stations_around(STATION_AROUND_RADIUS),
            options.places, options.max_stations)



"""Geo localizator objects"""
from functools import wraps
from math import sqrt, radians, sin, cos, atan2, pi
from urllib import quote_plus
from urllib import urlopen

import simplejson as json

from veliberator.models import StationInformation

EARTH_RADIUS = 6378137.0  # Earth radius in meters
EARTH_CIRCUMFERENCE = EARTH_RADIUS * 2 * pi
EARTH_METER_DEGREE = 360.0 / EARTH_CIRCUMFERENCE
EARTH_DEGREE_METER = EARTH_CIRCUMFERENCE / 360.0

global_geofinder_cache = {}


def cache_wrapper(method):
    @wraps(method)
    def cache(instance, around_radius):
        """Use a global cache for the results
        of stations around"""
        global global_geofinder_cache

        key_cache = (instance.lat, instance.lng, around_radius)
        if key_cache not in global_geofinder_cache.keys():
            global_geofinder_cache[key_cache] = method(instance, around_radius)

        return global_geofinder_cache[key_cache]
    return cache


def pythagor_distance(start, end):
    """Compute the distance between 2 points
    with Pythagor theorem"""
    l1 = float(start[0]) - float(end[0])
    l2 = float(start[1]) - float(end[1])
    return sqrt(pow(l1, 2) + pow(l2, 2)) * EARTH_DEGREE_METER


def haversine_distance(start, end):
    """Compute the distance between 2 points in meters
    with the haversine formula"""
    start_latt = radians(float(start[0]))
    start_long = radians(float(start[1]))
    end_latt = radians(float(end[0]))
    end_long = radians(float(end[1]))
    d_latt = end_latt - start_latt
    d_long = end_long - start_long
    a = sin(d_latt / 2) ** 2 + cos(start_latt) * \
        cos(end_latt) * sin(d_long / 2) ** 2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    return EARTH_RADIUS * c


class GeoFinderError(Exception):
    pass


class BaseGeoFinder(object):
    """Base GeoFinder object,
    for finding stations around"""
    lat = None
    lng = None

    def __init__(self, lat, lng):
        self.lat = lat
        self.lng = lng

    def compute_square_area(self, radius):
        """Compute the GPS coordonates of a square,
        around the position"""
        radius_degree = radius * EARTH_METER_DEGREE

        lat_orig = float(self.lat)
        lat_pos = lat_orig + radius_degree
        lat_neg = lat_orig - radius_degree

        lng_orig = float(self.lng)
        lng_pos = lng_orig + radius_degree
        lng_neg = lng_orig - radius_degree

        return (lat_neg, lat_orig, lat_pos), \
               (lng_neg, lng_orig, lng_pos)

    def compute_station_distances(self, stations,
                                  distance_function=haversine_distance):
        """Return the stations ordered by their distance to the
        origin, with a distance property added"""
        station_distances = {}

        for station in stations:
            if station.lat == self.lat and station.lng == self.lng:
                continue
            distance = distance_function((self.lat, self.lng),
                                         (station.lat, station.lng))
            station.distance = distance
            station_distances[station] = distance

        stations_sorted_by_distance = sorted(station_distances.iteritems(),
                                             key=lambda (k, v): (v, k))
        return [s for s, d in stations_sorted_by_distance]

    def get_stations_in_area(self, lats, lngs):
        """Make a query to find the station with
        lat and lng matching with the area"""
        return StationInformation.query.filter(
            (StationInformation.lat >= lats[0]) &
            (StationInformation.lat <= lats[2]) &
            (StationInformation.lng >= lngs[0]) &
            (StationInformation.lng <= lngs[2])).all()

    @cache_wrapper
    def get_stations_around(self, around_radius):
        """Find the stations around the lat and lng,
        sorted by proximity"""
        lats, lngs = self.compute_square_area(around_radius)
        stations = self.get_stations_in_area(lats, lngs)
        stations_distanced = self.compute_station_distances(stations)
        return stations_distanced


class StationGeoFinder(BaseGeoFinder):

    def __init__(self, station):
        """Init the object with a Station object"""
        self.lat = station.informations.lat
        self.lng = station.informations.lng


class AddressGeoFinder(BaseGeoFinder):

    def __init__(self, address):
        """Init the object with an address
        who will be geocomputed"""
        self.address = address
        informations = self.geocompute(address)

        self.lat = informations['geometry']['location']['lat']
        self.lng = informations['geometry']['location']['lng']
        self.precision = informations['types'][0]
        self.clean_address = informations['formatted_address']

    def geocompute(self, address):
        """Geocompute an address with GMap"""
        address = quote_plus(address)
        request = 'https://maps.googleapis.com/maps/api/geocode/json' \
                  '?address=%s&region=fr' % address
        data = json.loads(urlopen(request).read())

        if data['status'] != 'OK':
            raise GeoFinderError('Geocoding service error')
        return data['results'][0]



"""Models for database"""
from datetime import datetime

from elixir import Field
from elixir import Float
from elixir import Entity
from elixir import String
from elixir import Integer
from elixir import Boolean
from elixir import DateTime
from elixir import metadata
from elixir import setup_all
from elixir import create_all
from elixir import using_options

from veliberator.settings import DATABASE_URI, DATABASE_ECHO


class StationInformation(Entity):
    """Station information entity model"""
    id = Field(Integer, primary_key=True)

    address = Field(String(100))
    postal_code = Field(String(100))
    city = Field(String(100))

    lat = Field(Float)
    lng = Field(Float)

    opened = Field(Boolean)
    bonus = Field(Boolean)

    creation_date = Field(DateTime, default=datetime.now)
    update = Field(DateTime, default=datetime.now)

    using_options(order_by='id',
                  tablename='veliberator_station_information')

    def __repr__(self):
        return '<StationInformation "%s" (%s)>' % (self.id, self.address)

    @property
    def full_address(self):
        return '%s, %s %s' % (self.address,
                              self.postal_code,
                              self.city)


def db_connection(sqluri=DATABASE_URI, echo=DATABASE_ECHO):
    metadata.bind = sqluri
    metadata.bind.echo = echo
    setup_all()
    create_all()
